@@ -93,7 +93,7 @@ static
 arglist_t compegps_args[] = {
 	{"deficon", &option_icon, "Default icon name", 
 		NULL, ARGTYPE_STRING, ARG_NOMINMAX },
-	{"index", &option_index, "Index of route/track to write (if more the one in source)", 
+	{"index", &option_index, "Index of route/track to write (if more than one in source)", 
 		NULL, ARGTYPE_INT, "1", NULL},
 	{"radius", &option_radius, "Give points (waypoints/route points) a default radius (proximity)", 
 		NULL, ARGTYPE_FLOAT, "0", NULL},
@@ -359,7 +359,7 @@ compegps_data_read(void)
 		char *cin = buff;
 		char *ctail;
 		
-		line++;
+		if ((line++ == 0) && fin->unicode) cet_convert_init(CET_CHARSET_UTF8, 1);
 		cin = lrtrim(buff);
 		if (strlen(cin) == 0) continue;
 		
