@@ -1,7 +1,7 @@
 /*
     Describe vectors containing filter operations.
  
-    Copyright (C) 2002,2004,2005,2006 Robert Lipe, robertlipe@usa.net
+    Copyright (C) 2002,2004,2005,2006,2007 Robert Lipe, robertlipe@usa.net
 
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -22,6 +22,7 @@
 #include "defs.h"
 #include "filterdefs.h"
 #include "inifile.h"
+#include "gbversion.h"
 
 typedef struct {
 	filter_vecs_t *vec;
@@ -43,6 +44,8 @@ extern filter_vecs_t discard_vecs;
 extern filter_vecs_t nuke_vecs;
 extern filter_vecs_t interpolatefilt_vecs;
 extern filter_vecs_t transform_vecs;
+extern filter_vecs_t height_vecs;
+extern filter_vecs_t swapdata_vecs;
 
 static
 fl_vecs_t filter_vec_list[] = {
@@ -117,6 +120,17 @@ fl_vecs_t filter_vec_list[] = {
 		"transform",
 		"Transform waypoints into a route, tracks into routes, ..."
 	},
+	{
+		&height_vecs,
+		"height",
+		"Manipulate altitudes"
+	},
+	{
+		&swapdata_vecs,
+		"swap",
+		"Swap latitude and longitude of all loaded points"
+	},
+	
 #endif
         {
 		NULL,
@@ -145,7 +159,7 @@ find_filter_vec(char *const vecname, char **opts)
 		/* step 1: initialize by inifile or default values */
 		if (vec->vec->args) {
 			for (ap = vec->vec->args; ap->argstring; ap++) {
-				char *temp;
+				const char *temp;
 				
 				temp = inifile_readstr(global_opts.inifile, vec->name, ap->argstring);
 				if (temp == NULL) temp = inifile_readstr(global_opts.inifile, "Common filter settings", ap->argstring);
@@ -194,14 +208,29 @@ free_filter_vec( filter_vecs_t *fvec )
 	
 	if ( fvec->args ) {
 		for ( ap = fvec->args; ap->argstring; ap++) {
-			if (ap->argval && *ap->argval) {
-				xfree(*ap->argval);
-				*ap->argval = NULL;
+			if (ap->argvalptr) {
+				xfree(ap->argvalptr);
+				ap->argvalptr = *ap->argval = NULL;
 			}
 		}
 	}
 }
 
+void 
+init_filter_vecs(void)
+{
+	fl_vecs_t *vec = filter_vec_list;
+	while ( vec->vec ) {
+		arglist_t *ap;
+		if ( vec->vec->args ) {
+			for ( ap = vec->vec->args; ap->argstring; ap++ ) {
+				ap->argvalptr = NULL;
+			}
+		}
+		vec++;
+	}
+}
+
 void 
 exit_filter_vecs( void )
 {
@@ -213,7 +242,6 @@ exit_filter_vecs( void )
 		vec++;
 	}
 }
-		
 
 /*
  *  Display the available formats in a format that's easy for humans to
@@ -267,14 +295,25 @@ alpha (const void *a, const void *b)
         return case_ignore_strcmp(ap->desc , bp->desc);
 }
 
+static 
+void disp_help_url(const fl_vecs_t *vec, arglist_t *arg)
+{
+	printf("\t" WEB_DOC_DIR "/fmt_%s.html", vec->name);
+	if (arg) {
+		printf("#fmt_%s_o_%s",vec->name, arg->argstring);
+	}
+}
+
 static void
 disp_v1(const fl_vecs_t *vec)
 {
 	arglist_t *ap;
 	
+	disp_help_url(vec, NULL);
+	printf("\n");
 	for (ap = vec->vec->args; ap && ap->argstring; ap++) {
-		if ( !(ap->argtype & ARGTYPE_HIDDEN))
-			printf("option\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n",
+		if ( !(ap->argtype & ARGTYPE_HIDDEN)) {
+			printf("option\t%s\t%s\t%s\t%s\t%s\t%s\t%s",
 				vec->name, 
 				ap->argstring, 
 				ap->helpstring, 
@@ -282,6 +321,9 @@ disp_v1(const fl_vecs_t *vec)
 				ap->defaultvalue? ap->defaultvalue : "",
 				ap->minvalue? ap->minvalue : "",
 				ap->maxvalue? ap->maxvalue : "");
+			disp_help_url(vec, ap);
+			printf("\n");
+		}
 	}
 }
 
@@ -305,9 +347,12 @@ disp_filters(int version)
 	case 0:
 	case 1:
 		for (vec = filter_vec_list; vec->vec; vec++) {
-			printf("%s\t%s\n", vec->name, vec->desc);
-			if (version > 0)
+			if (version == 0) {
+				printf("%s\t%s\n", vec->name, vec->desc);
+			} else {
+				printf("%s\t%s", vec->name, vec->desc);
 				disp_v1(vec);
+			}
 		}
 		break;
 	default:
