@@ -59,6 +59,8 @@
 	    2007/07/07: Better support for new fields since V3 (postal code/street address/instruction)
 	    2008/01/09: Fix handling of option category (cat)
 	    2008/04/27: Add zero to checklist of "unknown bytes"
+	    2008/08/17: Add concept of route/track line colors
+	    2008/09/11: Make format 'pipeable' (cached writes using gbfile memapi)
 */
 
 #include <stdio.h>
@@ -108,10 +110,10 @@
 
 /*******************************************************************************/
 
-/* static char gdb_release[] = "$Revision: 1.65 $"; */
-static char gdb_release_date[] = "$Date: 2008/05/04 23:09:08 $";
+/* static char gdb_release[] = "$Revision: 1.67 $"; */
+static char gdb_release_date[] = "$Date: 2008/09/11 20:41:43 $";
 
-static gbfile *fin, *fout;
+static gbfile *fin, *fout, *ftmp;
 static int gdb_ver, gdb_category, gdb_via, gdb_roadbook;
 
 static queue wayptq_in, wayptq_out, wayptq_in_hidden;
@@ -531,7 +533,7 @@ read_waypoint(gt_waypt_classes_e *waypt_class_out)
 	}
 	GMSD_SET(display, display);
 	
-	FREAD_i32;				/* color/colour !not implemented! */
+	FREAD_i32;				/* color !not implemented! */
 	icon = FREAD_i32;
 	GMSD_SET(icon, icon);			/* icon */
 	FREAD_STR(buf);				/* city */
@@ -685,7 +687,8 @@ read_route(void)
 	int points, warnings, links, i;
 	char buf[128];
 	bounds bounds;
-	
+	int color_idx;
+
 	rte_ct++;
 	warnings = 0;
 
@@ -859,7 +862,8 @@ read_route(void)
 	else {
 		rte->rte_url = gdb_fread_strlist();
 		
-		FREAD(buf, 4);			/* ?????????????????????????????????? */
+		color_idx = FREAD_i32;
+		rte->line_color.bbggrr = gt_color_value(color_idx);
 		FREAD(buf, 1);			/* ?????????????????????????????????? */
 
 		rte->rte_desc = FREAD_CSTR;
@@ -885,7 +889,8 @@ read_track(void)
 	route_head *res;
 	int points, index;
 	char dummy;
-	
+	int color_idx;
+
 	trk_ct++;
 
 	res = route_head_alloc();
@@ -893,8 +898,9 @@ read_track(void)
 //	res->rte_num = trk_ct;
 
 	FREAD(&dummy, 1);		/* display - 1 byte */
-	FREAD_i32;			/* color -   1 dword */
-	
+	color_idx = FREAD_i32;		/* color -  1 dword */
+	res->line_color.bbggrr = gt_color_value(color_idx);
+
 	points = FREAD_i32;
 	
 	for (index = 0; index < points; index++)
@@ -940,9 +946,10 @@ read_track(void)
 /*******************************************************************************/
 
 static void
-init_reader(const char *fname)
+gdb_rd_init(const char *fname)
 {
 	fin = gbfopen_le(fname, "rb", MYNAME);
+	ftmp = gbfopen_le(NULL, "wb", MYNAME);
 	read_file_header();
 	/* VERSION DEPENDENT CODE */
 	if (gdb_ver >= GDB_VER_UTF8)
@@ -964,32 +971,39 @@ init_reader(const char *fname)
 }
 
 static void
-done_reader(void)
+gdb_rd_deinit(void)
 {
 	disp_summary(fin);
 	gdb_flush_waypt_queue(&wayptq_in);
 	gdb_flush_waypt_queue(&wayptq_in_hidden);
+	gbfclose(ftmp);
 	gbfclose(fin);
 }
 
 static void
 read_data(void)
 {
+	gbfile *fsave;
 	int incomplete = 0;	/* number of incomplete reads */
 
 	for (;;) {
 		int len, delta;
 		char typ, dump;
 		gt_waypt_classes_e wpt_class;
-		gbsize_t pos;
 		waypoint *wpt;
 		route_head *trk, *rte;
 		
 		len = FREAD_i32;
 		FREAD(&typ, 1);
-		pos = gbftell(fin);
-		
 		if (typ == 'V') break;	/* break the loop */
+
+		gbfrewind(ftmp);
+		gbfwrite(NULL, 0, 0, ftmp);	/* truncate */
+		gbfcopyfrom(ftmp, fin, len);
+		gbfrewind(ftmp);
+		
+		fsave = fin;			/* swap standard 'fin' with cached input */
+		fin = ftmp;
 		
 		dump = 1;
 		wpt_class = GDB_DEF_CLASS;
@@ -1019,7 +1033,9 @@ read_data(void)
 				break;
 		}
 		
-		delta = (pos + len) - gbftell(fin);
+		fin = fsave;
+		delta = len - gbftell(ftmp);
+
 		if (dump && delta) {
 			if (! incomplete++) {
 				warning(MYNAME ":==========================================\n");
@@ -1041,7 +1057,6 @@ read_data(void)
 			}
 			warning("\n");
 		}
-		gbfseek(fin, pos + len, SEEK_SET);
 	}
 	
 	
@@ -1210,7 +1225,7 @@ write_waypoint(
 		FWRITE_CSTR(wpt->description);
 	FWRITE_DBL(WAYPT_GET(wpt, proximity, unknown_alt), unknown_alt);	/* proximity */
 	FWRITE_i32(display);			/* display */
-	FWRITE_i32(0);				/* color (colour) */
+	FWRITE_i32(0);				/* color */
 	FWRITE_i32(icon);			/* icon */
 	FWRITE_CSTR(GMSD_GET(city, ""));	/* city */
 	FWRITE_CSTR(GMSD_GET(state, ""));	/* state */
@@ -1412,7 +1427,8 @@ write_route(const route_head *rte, const char *rte_name)
 	}
 	else /* if (gdb_ver >= GDB_VER_3) */ {
 		FWRITE_CSTR_LIST(rte->rte_url);
-		FWRITE_i32(0x0E);	/* color ??? */
+		/* "Magenta" (14) is MapSource default */
+		FWRITE_i32( (rte->line_color.bbggrr < 0) ? 14 : gt_color_index_by_rgb(rte->line_color.bbggrr) );
 		FWRITE_C(0);
 		FWRITE_CSTR(rte->rte_desc);
 	}
@@ -1426,7 +1442,8 @@ write_track(const route_head *trk, const char *trk_name)
 	
 	FWRITE_CSTR(trk_name);
 	FWRITE_C(0);
-	FWRITE_i32(0);
+	/* "Unknown" (0) is MapSource default */
+	FWRITE_i32(gt_color_index_by_rgb(trk->line_color.bbggrr));
 
 	FWRITE_i32(points);	/* total number of waypoints in waypoint list */
 	
@@ -1461,16 +1478,19 @@ write_track(const route_head *trk, const char *trk_name)
 /*-----------------------------------------------------------------------------*/
 
 static void
-finalize_item(const gbsize_t anchor)
+finalize_item(gbfile *origin, const char identifier)
 {
-	gbsize_t mark;
-	int len;
+	int len = gbftell(fout);
 	
-	mark = gbftell(fout);
-	len = mark - anchor;
-	gbfseek(fout, -(len + 5), SEEK_CUR);
-	FWRITE_i32(mark - anchor);
-	gbfseek(fout, len + 1, SEEK_CUR);
+	fout = origin;
+	gbfseek(ftmp, 0, SEEK_SET);
+	
+	FWRITE_i32(len);
+	FWRITE_C(identifier);
+	gbfcopyfrom(fout, ftmp, len);
+
+	gbfseek(ftmp, 0, SEEK_SET);	/* Truncate memory stream */
+	gbfwrite(NULL, 0, 0, ftmp);
 }
 
 /*-----------------------------------------------------------------------------*/
@@ -1492,7 +1512,7 @@ write_waypoint_cb(const waypoint *refpt)
 {
 	garmin_fs_t *gmsd;
 	waypoint *test;
-	gbsize_t anchor;
+	gbfile *fsave;
 	
 	/* do this when backup always happens in main */
 	
@@ -1513,9 +1533,8 @@ write_waypoint_cb(const waypoint *refpt)
 		gdb_check_waypt(wpt);
 		ENQUEUE_TAIL(&wayptq_out, &wpt->Q);
 		
-		FWRITE_i32(-1);
-		FWRITE_C('W');
-		anchor = gbftell(fout);
+		fsave = fout;
+		fout = ftmp;
 		
 		/* prepare the waypoint */
 		gmsd = GMSD_FIND(wpt);
@@ -1563,14 +1582,14 @@ write_waypoint_cb(const waypoint *refpt)
 		wpt->extra_data = (void *)name;
 		write_waypoint(wpt, name, gmsd, icon, display);
 
-		finalize_item(anchor);
+		finalize_item(fsave, 'W');
 	}
 }
 
 static void
 write_route_cb(const route_head *rte)
 {
-	gbsize_t anchor;
+	gbfile *fsave;
 	char *name;
 	char buf[32];
 	
@@ -1585,12 +1604,10 @@ write_route_cb(const route_head *rte)
 	
 	rte_ct++;	/* increase informational number of written routes */
 
-	FWRITE_i32(-1);
-	FWRITE_C('R');
-
-	anchor = gbftell(fout);
+	fsave = fout;
+	fout = ftmp;
 	write_route(rte, name);
-	finalize_item(anchor);
+	finalize_item(fsave, 'R');
 
 	xfree(name);
 }
@@ -1598,7 +1615,7 @@ write_route_cb(const route_head *rte)
 static void
 write_track_cb(const route_head *trk)
 {
-	gbsize_t anchor;
+	gbfile *fsave;
 	char *name;
 	char buf[32];
 	
@@ -1613,12 +1630,10 @@ write_track_cb(const route_head *trk)
 
 	trk_ct++;	/* increase informational number of written tracks */
 	
-	FWRITE_i32(-1);
-	FWRITE_C('T');
-
-	anchor = gbftell(fout);
+	fsave = fout;
+	fout = ftmp;
 	write_track(trk, name);
-	finalize_item(anchor);
+	finalize_item(fsave, 'T');
 
 	xfree(name);
 }
@@ -1626,9 +1641,10 @@ write_track_cb(const route_head *trk)
 /*-----------------------------------------------------------------------------*/
 
 static void
-init_writer(const char *fname)
+gdb_wr_init(const char *fname)
 {
 	fout = gbfopen_le(fname, "wb", MYNAME);
+	ftmp = gbfopen_le(NULL, "wb", MYNAME);
 
 	gdb_category = (gdb_opt_category) ? atoi(gdb_opt_category) : 0;
 	gdb_ver = (gdb_opt_ver && *gdb_opt_ver) ? atoi(gdb_opt_ver) : 0;
@@ -1658,12 +1674,13 @@ init_writer(const char *fname)
 }
 
 static void
-done_writer(void)
+gdb_wr_deinit(void)
 {
 	disp_summary(fout);
 	gdb_flush_waypt_queue(&wayptq_out);
 	mkshort_del_handle(&short_h);
 	gbfclose(fout);
+	gbfclose(ftmp);
 }
 
 static void
@@ -1719,10 +1736,10 @@ static arglist_t gdb_args[] = {
 ff_vecs_t gdb_vecs = {
 	ff_type_file,
 	FF_CAP_RW_ALL,
-	init_reader,	
-	init_writer,
-	done_reader,
-	done_writer,
+	gdb_rd_init,	
+	gdb_wr_init,
+	gdb_rd_deinit,
+	gdb_wr_deinit,
 	read_data,
 	write_data,
 	NULL, 
