@@ -26,12 +26,13 @@
 #include "garmin_fs.h"
 #include "garmin_tables.h"
 #include "jeeps/gpsmath.h"
-#include "strptime.h"
 #include <string.h>
 #include <time.h>
 
 #define MYNAME "unicsv"
 
+#define UNICSV_GC_READY
+
 /* "UNICSV_FIELD_SEP" and "UNICSV_LINE_SEP" are only used by the writer */
 
 #define UNICSV_FIELD_SEP	","
@@ -88,13 +89,31 @@ typedef enum {
 	fld_sec,
 	fld_ns,
 	fld_ew,
+
 	fld_garmin_city,
 	fld_garmin_postal_code,
 	fld_garmin_state,
 	fld_garmin_country,
 	fld_garmin_addr,
 	fld_garmin_phone_nr,
+	fld_garmin_phone_nr2,
+	fld_garmin_fax_nr,
+	fld_garmin_email,
 	fld_garmin_facility,
+#ifdef UNICSV_GC_READY
+	fld_gc_id,
+	fld_gc_type,
+	fld_gc_container,
+	fld_gc_terr,
+	fld_gc_diff,
+	fld_gc_is_archived,
+	fld_gc_is_available,
+	fld_gc_exported,
+	fld_gc_last_found,
+	fld_gc_placer,
+	fld_gc_placer_id,
+	fld_gc_hint,
+#endif
 	fld_terminator
 } field_e;
 
@@ -107,7 +126,7 @@ typedef enum {
 #define unicsv_unknown	1e25
 
 typedef struct {
-	char *name;
+	const char *name;
 	field_e type;
 	gbuint32 options;
 } field_t;
@@ -201,8 +220,26 @@ static field_t fields_def[] = {
 	{ "post",	fld_garmin_postal_code, STR_ANY },
 	{ "zip",	fld_garmin_postal_code, STR_ANY },
 	{ "phone",	fld_garmin_phone_nr, STR_ANY },
+	{ "phone2",	fld_garmin_phone_nr2, STR_ANY },
+	{ "fax",	fld_garmin_fax_nr, STR_ANY },
+	{ "email",	fld_garmin_email, STR_ANY },
 	{ "state",	fld_garmin_state, STR_ANY },
 	{ "faci",	fld_garmin_facility, STR_ANY },
+#ifdef UNICSV_GC_READY
+	/* geocache details */
+	{ "gcid",	fld_gc_id, STR_ANY },
+	{ "type",	fld_gc_type, STR_ANY },
+	{ "cont",	fld_gc_container, STR_ANY },
+	{ "terr",	fld_gc_terr, STR_ANY },
+	{ "diff",	fld_gc_diff, STR_ANY },
+	{ "arch",	fld_gc_is_archived, STR_ANY },
+	{ "avail",	fld_gc_is_available, STR_ANY },
+	{ "exported",	fld_gc_exported, STR_ANY },
+	{ "found",	fld_gc_last_found, STR_ANY },
+	{ "placer",	fld_gc_placer, STR_ANY },
+	{ "placer_id",	fld_gc_placer_id, STR_ANY },
+	{ "hint",	fld_gc_hint, STR_ANY },
+#endif
 	{ NULL,		fld_terminator, 0 }
 };
 
@@ -210,24 +247,31 @@ static field_e *unicsv_fields_tab;
 static int unicsv_fields_tab_ct;
 static double unicsv_altscale, unicsv_depthscale, unicsv_proximityscale
 ;
-static char *unicsv_fieldsep;
+static const char *unicsv_fieldsep;
 static gbfile *fin, *fout;
 static gpsdata_type unicsv_data_type;
 static route_head *unicsv_track, *unicsv_route;
-static unsigned long long unicsv_outp_flags;
+static char unicsv_outp_flags[(fld_terminator + 8) / 8];
 static grid_type unicsv_grid_idx;
 static int unicsv_datum_idx;
-static char *opt_datum, *opt_grid, *opt_utc;
+static char *opt_datum, *opt_grid, *opt_utc, *opt_filename, *opt_format, *opt_prec;
 static int unicsv_waypt_ct;
 static char unicsv_detect;
+static int llprec;
 
-static arglist_t unicsv_args[] = { 
+static arglist_t unicsv_args[] = {
 	{"datum", &opt_datum, "GPS datum (def. WGS 84)",
-		"WGS 84", ARGTYPE_STRING, ARG_NOMINMAX}, 
+		"WGS 84", ARGTYPE_STRING, ARG_NOMINMAX},
 	{"grid",  &opt_grid,  "Write position using this grid.",
 		NULL, ARGTYPE_STRING, ARG_NOMINMAX},
-	{"utc",   &opt_utc,   "Write timestamps with offset x to UTC time", 
+	{"utc",   &opt_utc,   "Write timestamps with offset x to UTC time",
 		NULL, ARGTYPE_INT, "-23", "+23"},
+	{"format", &opt_format,   "Write name(s) of format(s) from input session(s)",
+		NULL, ARGTYPE_BOOL, ARG_NOMINMAX},
+	{"filename", &opt_filename,   "Write filename(s) from input session(s)",
+		NULL, ARGTYPE_BOOL, ARG_NOMINMAX},
+	{"prec", &opt_prec,   "Precision of numerical coordinates (no grid set)",
+		"6", ARGTYPE_INT | ARGTYPE_HIDDEN, "0", "15"},
 	ARG_TERMINATOR };
 
 
@@ -241,7 +285,7 @@ static int
 unicsv_strrcmp(const char *s1, const char *s2)
 {
 	int l1, l2;
-	
+
 	l1 = strlen(s1);
 	l2 = strlen(s2);
 	if ((l1 - l2) >= 0)
@@ -251,13 +295,42 @@ unicsv_strrcmp(const char *s1, const char *s2)
 }
 
 static int
+unicsv_parse_gc_id(const char *str)
+{
+	int res = 0;
+
+	if (str && (str[0] == 'G') && (str[1] == 'C')) {
+		int base;
+		char cx;
+
+		str += 2;
+		if (strlen(str) > 4) base = 31;
+		else base = (*str < 'G') ? 16 : 31;
+
+		while ((cx = *str++)) {
+			int num;
+
+			if ((cx >= '0') && (cx <= '9')) num = cx - '0';
+			else if ((cx >= 'A') && (cx <= 'Z')) num = cx - 'A' + 10;
+			else break;
+
+			res = (res * base) + num;
+		}
+		if (base == 31) res -= 411120;
+	}
+	return res;
+}
+
+// static int unicsv_parse_time(const char *str, int *msec, time_t *date);
+
+static time_t
 unicsv_parse_date(const char *str, int *consumed)
 {
 	int p1, p2, p3, ct;
 	char sep[2];
 	struct tm tm;
 	int lconsumed = 0;
-	
+
 	memset(&tm, 0, sizeof(tm));
 	ct = sscanf(str, "%d%1[-.//]%d%1[-.//]%d%n", &p1, sep, &p2, sep, &p3, &lconsumed);
 	if (consumed && lconsumed) {
@@ -270,7 +343,7 @@ unicsv_parse_date(const char *str, int *consumed)
 		}
 		fatal(MYNAME ": Could not parse date string (%s).\n", str);
 	}
-	
+
 	if ((p1 > 99) || (sep[0] == '-')) { /* Y-M-D (iso like) */
 		tm.tm_year = p1;
 		tm.tm_mon = p2;
@@ -298,21 +371,21 @@ unicsv_parse_date(const char *str, int *consumed)
 		}
 		fatal(MYNAME ": Could not parse date string (%s).\n", str);
 	}
-	
+
 	tm.tm_year -= 1900;
 	tm.tm_mon -= 1;
-	
+
 	return mkgmtime(&tm);
 }
 
-static int
-unicsv_parse_time(const char *str, int *msec, int *date)
+static time_t
+unicsv_parse_time(const char *str, int *msec, time_t *date)
 {
 	int hour, min, ct, sec;
 	int consumed = 0;
 	double ms;
-	char sep[1];
-	int ldate;
+	char sep[2];
+	time_t ldate;
 
 	/* If we have somethine we're pretty sure is a date, parse that
 	 * first, skip over it, and pass that back to the caller)
@@ -324,7 +397,7 @@ unicsv_parse_time(const char *str, int *msec, int *date)
 			*date = ldate;
 		}
 	}
-	
+
 	ct = sscanf(str, "%d%1[.://]%d%1[.://]%d%lf", &hour, sep, &min, sep, &sec, &ms);
 	is_fatal(ct < 5, MYNAME ": Could not parse time string (%s).\n", str);
 	if (ct == 6) {
@@ -339,6 +412,35 @@ unicsv_parse_time(const char *str, int *msec, int *date)
 	return ((hour * SECONDS_PER_HOUR) + (min * 60) + (int)sec);
 }
 
+#ifdef UNICSV_GC_READY
+static status_type
+unicsv_parse_status(const char *str)
+{
+	if ((case_ignore_strcmp(str, "true") == 0) ||
+	    (case_ignore_strcmp(str, "yes") == 0) ||
+	    (*str == '1')) return status_true;
+	else if ((case_ignore_strcmp(str, "false") == 0) ||
+		 (case_ignore_strcmp(str, "no") == 0) ||
+		 (*str == '0')) return status_false;
+	else return status_unknown;
+}
+#endif
+
+#ifdef UNICSV_GC_READY
+static time_t
+unicsv_adjust_time(const time_t time, time_t *date)
+{
+	time_t res = time;
+	if (date) res += *date;
+	if (opt_utc) res += atoi(opt_utc) * SECONDS_PER_HOUR;
+	else {
+		struct tm tm = *gmtime(&res);
+		res = mklocaltime(&tm);
+	}
+	return res;
+}
+#endif
+
 static char
 unicsv_compare_fields(char *s, const field_t *f)
 {
@@ -399,10 +501,10 @@ unicsv_fondle_header(char *ibuf)
 	int i, column;
 	const cet_cs_vec_t *ascii = &cet_cs_vec_ansi_x3_4_1968;	/* us-ascii */
 
-	/* Convert the entire header to lower case for convenience. 
+	/* Convert the entire header to lower case for convenience.
 	 * If we see a tab in that header, we decree it to be tabsep.
 	 */
-	unicsv_fieldsep = ",";	
+	unicsv_fieldsep = ",";
 	for (s = ibuf; *s; s++) {
 		if (*s == '\t') {
 			unicsv_fieldsep = "\t";
@@ -421,7 +523,7 @@ unicsv_fondle_header(char *ibuf)
 	for (s = ibuf; *s; s++) {
 		*s = tolower(*s);
 	}
-	
+
 	/* convert the header line into native ascii */
 	if (global_opts.charset != ascii) {
 		buf = cet_str_any_to_any(ibuf, global_opts.charset, ascii);
@@ -430,14 +532,14 @@ unicsv_fondle_header(char *ibuf)
 
 	column = -1;
 	while ((s = csv_lineparse(ibuf, unicsv_fieldsep, "\"", 0))) {
-		
+
 		field_t *f = &fields_def[0];
-		
+
 		ibuf = NULL;
 		column++;
 		unicsv_fields_tab_ct++;
 		s = lrtrim(s);
-		
+
 		if (column % 4 == 0) {
 			int sz = (column + 4) * sizeof(*unicsv_fields_tab);
 			if (column == 0) unicsv_fields_tab = xmalloc(sz);
@@ -454,7 +556,7 @@ unicsv_fondle_header(char *ibuf)
 		}
 		if ((! f->name) && global_opts.debug_level)
 				warning(MYNAME ": Unhandled column \"%s\".\n", s);
-		
+
 		/* handle some special items */
 		if (f->type == fld_altitude) {
 			if (UNICSV_CONTAINS("ft") || UNICSV_CONTAINS("feet")) {
@@ -486,7 +588,7 @@ unicsv_rd_init(const char *fname)
 	unicsv_altscale = 1.0;
 	unicsv_depthscale = 1.0;
 	unicsv_proximityscale = 1.0;
-	
+
 	unicsv_fields_tab = NULL;
 	unicsv_fields_tab_ct = 0;
 	unicsv_data_type = global_opts.objective;
@@ -501,6 +603,7 @@ unicsv_rd_init(const char *fname)
 		unicsv_fondle_header(c);
 	else
 		unicsv_fieldsep = NULL;
+	if (fin->unicode) cet_convert_init(CET_CHARSET_UTF8, 1);
 }
 
 static void
@@ -526,7 +629,8 @@ unicsv_parse_one_line(char *ibuf)
 	double swiss_easting = unicsv_unknown;
 	double swiss_northing = unicsv_unknown;
 	int checked = 0;
-	int date = -1, time = -1, msec = -1;
+	time_t date = -1, time = -1;
+	int msec = -1;
 	char is_localtime = 0;
 	garmin_fs_t *gmsd;
 	double d;
@@ -534,7 +638,9 @@ unicsv_parse_one_line(char *ibuf)
 	int src_datum = unicsv_datum_idx;
 	int ns = 1;
 	int ew = 1;
-
+#ifdef UNICSV_GC_READY
+	geocache_data *gc_data = NULL;
+#endif
 	wpt = waypt_new();
 	wpt->latitude = unicsv_unknown;
 	wpt->longitude = unicsv_unknown;
@@ -542,7 +648,7 @@ unicsv_parse_one_line(char *ibuf)
 
 	column = -1;
 	while ((s = csv_lineparse(ibuf, unicsv_fieldsep, "\"", 0))) {
-	
+
 		if (column > unicsv_fields_tab_ct) break;	/* ignore extra fields on line */
 
 		ibuf = NULL;
@@ -556,6 +662,7 @@ unicsv_parse_one_line(char *ibuf)
 
 		case fld_time:
 		case fld_date:
+		case fld_datetime:
 			/* switch column type if it looks like an iso time string */
 			if (strchr(s, 'T'))
 				unicsv_fields_tab[column] = fld_iso_time;
@@ -565,21 +672,21 @@ unicsv_parse_one_line(char *ibuf)
 
 
 		switch(unicsv_fields_tab[column]) {
-		
+
 		case fld_latitude:
 			human_to_dec( s, &wpt->latitude, &wpt->longitude, 1 );
 			wpt->latitude = wpt->latitude * ns;
 			break;
-			
+
 		case fld_longitude:
 			human_to_dec( s, &wpt->latitude, &wpt->longitude, 2 );
 			wpt->longitude = wpt->longitude * ew;
 			break;
-			
+
 		case fld_shortname:
 			wpt->shortname = xstrdup(s);
 			break;
-			
+
 		case fld_description:
 			wpt->description = xstrdup(s);
 			break;
@@ -602,19 +709,19 @@ unicsv_parse_one_line(char *ibuf)
 		case fld_utm_zone:
 			utm_zone = atoi(s);
 			break;
-			
+
 		case fld_utm_easting:
 			utm_easting = atof(s);
 			break;
-			
+
 		case fld_utm_northing:
 			utm_northing = atof(s);
 			break;
-			
+
 		case fld_utm_zone_char:
 			utm_zc = toupper(s[0]);
 			break;
-			
+
 		case fld_utm:
 			parse_coordinates(s, unicsv_datum_idx, grid_utm,
 				&wpt->latitude, &wpt->longitude, MYNAME);
@@ -630,7 +737,7 @@ unicsv_parse_one_line(char *ibuf)
 			   don't convert a second time */
 			src_datum = DATUM_WGS84;
 			break;
-			
+
 		case fld_bng_zone:
 			strncpy(bng_zone, s, sizeof(bng_zone));
 			strupper(bng_zone);
@@ -643,7 +750,7 @@ unicsv_parse_one_line(char *ibuf)
 		case fld_bng_easting:
 			bng_easting = atof(s);
 			break;
-			
+
 		case fld_swiss:
 			parse_coordinates(s, DATUM_WGS84, grid_swiss,
 				&wpt->latitude, &wpt->longitude, MYNAME);
@@ -721,7 +828,7 @@ unicsv_parse_one_line(char *ibuf)
 			WAYPT_SET(wpt, course, atof(s));
 			if (unicsv_detect) unicsv_data_type = trkdata;
 			break;
-			
+
 		case fld_temperature:
 			d = atof(s);
 			if (fabs(d) < 999999) WAYPT_SET(wpt, temperature, d);
@@ -751,12 +858,12 @@ unicsv_parse_one_line(char *ibuf)
 			if (parse_distance(s, &d, unicsv_depthscale, MYNAME))
 				WAYPT_SET(wpt, depth, d);
 			break;
-			
+
 		case fld_symbol:
 			wpt->icon_descr = xstrdup(s);
 			wpt->wpt_flags.icon_descr_is_dynamic = 1;
 			break;
-		
+
 		case fld_iso_time:
 			is_localtime = 2;	/* fix result */
 			wpt->creation_time = xml_parse_time(s, &wpt->microseconds);
@@ -769,13 +876,13 @@ unicsv_parse_one_line(char *ibuf)
 			}
 			break;
 
-		case fld_date: 
+		case fld_date:
 			if ((is_localtime < 2) && (date < 0)) {
 				date = unicsv_parse_date(s, NULL);
 				is_localtime = 1;
 			}
 			break;
-			
+
 		case fld_year:
 			ymd.tm_year = atoi(s);
 			break;
@@ -801,22 +908,31 @@ unicsv_parse_one_line(char *ibuf)
 			break;
 
 		case fld_datetime:
-			/* not implemented */
+			if ((is_localtime < 2) && (date < 0) && (time < 0)) {
+				time = unicsv_parse_time(s, &msec, &date);
+				is_localtime = 1;
+			}
 			break;
+
 		case fld_ns:
 			ns = tolower(s[0]) == 'n' ? 1 : -1;
 			wpt->latitude *= ns;
 			break;
+
 		case fld_ew:
 			ew = tolower(s[0]) == 'e' ? 1 : -1;
 			wpt->longitude *= ew;
 			break;
+
 		case fld_garmin_city:
 		case fld_garmin_postal_code:
 		case fld_garmin_state:
 		case fld_garmin_country:
 		case fld_garmin_addr:
 		case fld_garmin_phone_nr:
+		case fld_garmin_phone_nr2:
+		case fld_garmin_fax_nr:
+		case fld_garmin_email:
 		case fld_garmin_facility:
 			gmsd = GMSD_FIND(wpt);
 			if (! gmsd) {
@@ -830,11 +946,62 @@ unicsv_parse_one_line(char *ibuf)
 			case fld_garmin_country: GMSD_SETSTR(country, s); break;
 			case fld_garmin_addr: GMSD_SETSTR(addr, s); break;
 			case fld_garmin_phone_nr: GMSD_SETSTR(phone_nr, s); break;
+			case fld_garmin_phone_nr2: GMSD_SETSTR(phone_nr2, s); break;
+			case fld_garmin_fax_nr: GMSD_SETSTR(fax_nr, s); break;
+			case fld_garmin_email: GMSD_SETSTR(email, s); break;
 			case fld_garmin_facility: GMSD_SETSTR(facility, s); break;
 			default: break;
 			}
 			break;
+#ifdef UNICSV_GC_READY
+		case fld_gc_id:
+		case fld_gc_type:
+		case fld_gc_container:
+		case fld_gc_terr:
+		case fld_gc_diff:
+		case fld_gc_is_archived:
+		case fld_gc_is_available:
+		case fld_gc_exported:
+		case fld_gc_last_found:
+		case fld_gc_placer:
+		case fld_gc_placer_id:
+		case fld_gc_hint:
 
+			gc_data = waypt_alloc_gc_data(wpt);
+
+			switch(unicsv_fields_tab[column]) {
+
+			case fld_gc_id:
+				gc_data->id = atoi(s);
+				if (gc_data->id == 0) gc_data->id = unicsv_parse_gc_id(s);
+				break;
+			case fld_gc_type: gc_data->type = gs_mktype(s); break;
+			case fld_gc_container: gc_data->container = gs_mkcont(s); break;
+			case fld_gc_terr: gc_data->terr = atof(s) * 10; break;
+			case fld_gc_diff: gc_data->diff = atof(s) * 10; break;
+			case fld_gc_is_archived: gc_data->is_archived = unicsv_parse_status(s); break;
+			case fld_gc_is_available: gc_data->is_available = unicsv_parse_status(s); break;
+			case fld_gc_exported: {
+				time_t time, date; int msec;
+				time = unicsv_parse_time(s, &msec, &date);
+				if (date || time) gc_data->exported = unicsv_adjust_time(time, &date);
+				}
+				break;
+			case fld_gc_last_found: {
+				time_t time, date;
+				int msec;
+				time = unicsv_parse_time(s, &msec, &date);
+				if (date || time) gc_data->last_found = unicsv_adjust_time(time, &date);
+				}
+				break;
+			case fld_gc_placer: gc_data->placer = xstrdup(s); break;
+			case fld_gc_placer_id: gc_data->placer_id = atoi(s); break;
+			case fld_gc_hint: gc_data->hint = xstrdup(s); break;
+
+			default: break;
+			}
+			break;
+#endif
 		case fld_terminator: /* dummy */
 			checked--;
 			break;
@@ -845,11 +1012,11 @@ unicsv_parse_one_line(char *ibuf)
 		waypt_free(wpt);
 		return;
 	}
-	
+
 	if (is_localtime < 2) {	/* not fixed */
 		if ((time >= 0) && (date >= 0)) {
 			time_t t = date + time;
-			
+
 			if (is_localtime) {
 				struct tm tm;
 				tm = *gmtime(&t);
@@ -871,10 +1038,10 @@ unicsv_parse_one_line(char *ibuf)
 				else ymd.tm_year += 1900;
 			}
 			ymd.tm_year -= 1900;
-			
+
 			if (ymd.tm_mon == 0) ymd.tm_mon = 1;
 			if (ymd.tm_mday == 0) ymd.tm_mday = 1;
-			
+
 			ymd.tm_mon--;
 			if (opt_utc)
 				wpt->creation_time = mkgmtime(&ymd);
@@ -890,11 +1057,11 @@ unicsv_parse_one_line(char *ibuf)
 
 		if (msec >= 0)
 			wpt->microseconds = msec;
-		
+
 		if (opt_utc)
 			wpt->creation_time += atoi(opt_utc) * SECONDS_PER_HOUR;
 	}
-	
+
 	/* utm/bng/swiss can be optional */
 
 	if ((wpt->latitude == unicsv_unknown) && (wpt->longitude == unicsv_unknown)) {
@@ -911,16 +1078,16 @@ unicsv_parse_one_line(char *ibuf)
 			src_datum = DATUM_WGS84;	/* don't convert afterwards */
 		}
 		else if ((swiss_easting != unicsv_unknown) && (swiss_northing != unicsv_unknown)) {
-			GPS_Math_CH1903_NGEN_To_WGS84(swiss_easting, swiss_northing,
+			GPS_Math_Swiss_EN_To_WGS84(swiss_easting, swiss_northing,
 				&wpt->latitude, &wpt->longitude);
 			src_datum = DATUM_WGS84;	/* don't convert afterwards */
 		}
 	}
-	
-	if ((src_datum != DATUM_WGS84) && 
+
+	if ((src_datum != DATUM_WGS84) &&
 	    (wpt->latitude != unicsv_unknown) && (wpt->longitude != unicsv_unknown)) {
 		double alt;
-		GPS_Math_Known_Datum_To_WGS84_M(wpt->latitude, wpt->longitude, (double) 0.0, 
+		GPS_Math_Known_Datum_To_WGS84_M(wpt->latitude, wpt->longitude, (double) 0.0,
 			&wpt->latitude, &wpt->longitude, &alt, src_datum);
 	}
 
@@ -944,13 +1111,13 @@ unicsv_parse_one_line(char *ibuf)
 	}
 }
 
-static void 
+static void
 unicsv_rd(void)
 {
 	char *buff;
 
 	if (unicsv_fieldsep == NULL) return;
-	
+
 	while ((buff = gbfgetstr(fin))) {
 	    	buff = lrtrim(buff);
 		if ((*buff == '\0') || (*buff == '#')) continue;
@@ -975,7 +1142,7 @@ unicsv_print_str(const char *str)
 {
 	if (str && *str) {
 		char *cout, *cx;
-		
+
 		cout = strenquote(str, UNICSV_QUOT_CHAR);
 
 		while ((cx = strstr(cout, "\r\n"))) {
@@ -991,17 +1158,38 @@ unicsv_print_str(const char *str)
 			*cx++ = ',';
 			lrtrim(cx);
 		}
-		
+
 		gbfprintf(fout, "%s%s", unicsv_fieldsep, cout);
 		xfree(cout);
 	}
 	else gbfputs(unicsv_fieldsep, fout);
 }
 
-#define BIT_OF(a) (1ULL << a)
-#define FIELD_USED(a) (unicsv_outp_flags & (1ULL << a))
+#ifdef UNICSV_GC_READY
+static void
+unicsv_print_data_time(const time_t atime)
+{
+	struct tm tm;
+	time_t time = atime;
+	char buf[32] = "";
 
-static void 
+	if (time) {
+		if (opt_utc) {
+			time += atoi(opt_utc) * SECONDS_PER_HOUR;
+			tm = *gmtime(&time);
+		}
+		else tm = *localtime(&time);
+		snprintf(buf, sizeof(buf), "%04d/%02d/%02d %02d:%02d:%02d",
+			tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday,
+			tm.tm_hour, tm.tm_min, tm.tm_sec);
+	}
+	unicsv_print_str(buf);
+}
+#endif
+
+#define FIELD_USED(a) (gb_getbit(&unicsv_outp_flags, a))
+
+static void
 unicsv_waypt_enum_cb(const waypoint *wpt)
 {
 	char *shortname;
@@ -1009,62 +1197,86 @@ unicsv_waypt_enum_cb(const waypoint *wpt)
 
 	shortname = (wpt->shortname) ? wpt->shortname : "";
 	gmsd = GMSD_FIND(wpt);
-	
-	if (*shortname) unicsv_outp_flags |= BIT_OF(fld_shortname);
-	if (wpt->altitude != unknown_alt) unicsv_outp_flags |= BIT_OF(fld_altitude);
-	if (wpt->icon_descr && *wpt->icon_descr) unicsv_outp_flags |= BIT_OF(fld_symbol);
+
+	if (*shortname) gb_setbit(&unicsv_outp_flags, fld_shortname);
+	if (wpt->altitude != unknown_alt) gb_setbit(&unicsv_outp_flags, fld_altitude);
+	if (wpt->icon_descr && *wpt->icon_descr) gb_setbit(&unicsv_outp_flags, fld_symbol);
 	if (wpt->description && *wpt->description && (strcmp(shortname, wpt->description) != 0))
-		unicsv_outp_flags |= BIT_OF(fld_description);
+		gb_setbit(&unicsv_outp_flags, fld_description);
 	if (wpt->notes && *wpt->notes && (strcmp(shortname, wpt->notes) != 0)) {
 		if ((! wpt->description) || (strcmp(wpt->description, wpt->notes) != 0))
-		    	unicsv_outp_flags |= BIT_OF(fld_notes);
+		    	gb_setbit(&unicsv_outp_flags, fld_notes);
 	}
-	if (wpt->url && *wpt->url) unicsv_outp_flags |= BIT_OF(fld_url);
+	if (wpt->url && *wpt->url) gb_setbit(&unicsv_outp_flags, fld_url);
 	if (wpt->creation_time != 0) {
-		unicsv_outp_flags |= BIT_OF(fld_time);
+		gb_setbit(&unicsv_outp_flags, fld_time);
 		if (wpt->creation_time >= SECONDS_PER_DAY)
-			unicsv_outp_flags |= BIT_OF(fld_date);
+			gb_setbit(&unicsv_outp_flags, fld_date);
 	}
 
-	if (wpt->fix != fix_unknown) unicsv_outp_flags |= BIT_OF(fld_fix);
-	if (wpt->vdop > 0) unicsv_outp_flags |= BIT_OF(fld_vdop);
-	if (wpt->hdop > 0) unicsv_outp_flags |= BIT_OF(fld_hdop);
-	if (wpt->pdop > 0) unicsv_outp_flags |= BIT_OF(fld_pdop);
-	if (wpt->sat > 0) unicsv_outp_flags |= BIT_OF(fld_sat);
-	if (wpt->heartrate != 0) unicsv_outp_flags |= BIT_OF(fld_heartrate);
-	if (wpt->cadence != 0) unicsv_outp_flags |= BIT_OF(fld_cadence);
+	if (wpt->fix != fix_unknown) gb_setbit(&unicsv_outp_flags, fld_fix);
+	if (wpt->vdop > 0) gb_setbit(&unicsv_outp_flags, fld_vdop);
+	if (wpt->hdop > 0) gb_setbit(&unicsv_outp_flags, fld_hdop);
+	if (wpt->pdop > 0) gb_setbit(&unicsv_outp_flags, fld_pdop);
+	if (wpt->sat > 0) gb_setbit(&unicsv_outp_flags, fld_sat);
+	if (wpt->heartrate != 0) gb_setbit(&unicsv_outp_flags, fld_heartrate);
+	if (wpt->cadence != 0) gb_setbit(&unicsv_outp_flags, fld_cadence);
 
 	/* "flagged" waypoint members */
-	if WAYPT_HAS(wpt, course) unicsv_outp_flags |= BIT_OF(fld_course);
-	if WAYPT_HAS(wpt, depth) unicsv_outp_flags |= BIT_OF(fld_depth);
-	if WAYPT_HAS(wpt, speed) unicsv_outp_flags |= BIT_OF(fld_speed);
-	if WAYPT_HAS(wpt, proximity) unicsv_outp_flags |= BIT_OF(fld_proximity);
-	if WAYPT_HAS(wpt, temperature) unicsv_outp_flags |= BIT_OF(fld_temperature);
-	
+	if WAYPT_HAS(wpt, course) gb_setbit(&unicsv_outp_flags, fld_course);
+	if WAYPT_HAS(wpt, depth) gb_setbit(&unicsv_outp_flags, fld_depth);
+	if WAYPT_HAS(wpt, speed) gb_setbit(&unicsv_outp_flags, fld_speed);
+	if WAYPT_HAS(wpt, proximity) gb_setbit(&unicsv_outp_flags, fld_proximity);
+	if WAYPT_HAS(wpt, temperature) gb_setbit(&unicsv_outp_flags, fld_temperature);
+
 	if (gmsd) {
-		if GMSD_HAS(addr) unicsv_outp_flags |= BIT_OF(fld_garmin_addr);
-		if GMSD_HAS(city) unicsv_outp_flags |= BIT_OF(fld_garmin_city);
-		if GMSD_HAS(country) unicsv_outp_flags |= BIT_OF(fld_garmin_country);
-		if GMSD_HAS(phone_nr) unicsv_outp_flags |= BIT_OF(fld_garmin_phone_nr);
-		if GMSD_HAS(postal_code) unicsv_outp_flags |= BIT_OF(fld_garmin_postal_code);
-		if GMSD_HAS(state) unicsv_outp_flags |= BIT_OF(fld_garmin_state);
-		if GMSD_HAS(facility) unicsv_outp_flags |= BIT_OF(fld_garmin_facility);
+		if GMSD_HAS(addr) gb_setbit(&unicsv_outp_flags, fld_garmin_addr);
+		if GMSD_HAS(city) gb_setbit(&unicsv_outp_flags, fld_garmin_city);
+		if GMSD_HAS(country) gb_setbit(&unicsv_outp_flags, fld_garmin_country);
+		if GMSD_HAS(phone_nr) gb_setbit(&unicsv_outp_flags, fld_garmin_phone_nr);
+		if GMSD_HAS(phone_nr2) gb_setbit(&unicsv_outp_flags, fld_garmin_phone_nr2);
+		if GMSD_HAS(fax_nr) gb_setbit(&unicsv_outp_flags, fld_garmin_fax_nr);
+		if GMSD_HAS(email) gb_setbit(&unicsv_outp_flags, fld_garmin_email);
+		if GMSD_HAS(postal_code) gb_setbit(&unicsv_outp_flags, fld_garmin_postal_code);
+		if GMSD_HAS(state) gb_setbit(&unicsv_outp_flags, fld_garmin_state);
+		if GMSD_HAS(facility) gb_setbit(&unicsv_outp_flags, fld_garmin_facility);
 	}
+
+#ifdef UNICSV_GC_READY
+	if (! waypt_empty_gc_data(wpt)) {
+		const geocache_data *gc_data = wpt->gc_data;
+
+		if (gc_data->id) gb_setbit(&unicsv_outp_flags, fld_gc_id);
+		if (gc_data->type) gb_setbit(&unicsv_outp_flags, fld_gc_type);
+		if (gc_data->container) gb_setbit(&unicsv_outp_flags, fld_gc_container);
+		if (gc_data->terr) gb_setbit(&unicsv_outp_flags, fld_gc_terr);
+		if (gc_data->diff) gb_setbit(&unicsv_outp_flags, fld_gc_diff);
+		if (gc_data->is_archived) gb_setbit(&unicsv_outp_flags, fld_gc_is_archived);
+		if (gc_data->is_available) gb_setbit(&unicsv_outp_flags, fld_gc_is_available);
+		if (gc_data->exported) gb_setbit(&unicsv_outp_flags, fld_gc_exported);
+		if (gc_data->last_found) gb_setbit(&unicsv_outp_flags, fld_gc_last_found);
+		if (gc_data->placer && *gc_data->placer) gb_setbit(&unicsv_outp_flags, fld_gc_placer);
+		if (gc_data->placer_id) gb_setbit(&unicsv_outp_flags, fld_gc_placer_id);
+		if (gc_data->hint && *gc_data->hint) gb_setbit(&unicsv_outp_flags, fld_gc_hint);
+	}
+#endif
 }
 
-static void 
+static void
 unicsv_waypt_disp_cb(const waypoint *wpt)
 {
 	double lat, lon, alt;
 	char *cout = NULL;
 	char *shortname;
 	garmin_fs_t *gmsd;
-
+#ifdef UNICSV_GC_READY
+	const geocache_data *gc_data = NULL;
+#endif
 	unicsv_waypt_ct++;
 
 	shortname = (wpt->shortname) ? wpt->shortname : "";
 	gmsd = GMSD_FIND(wpt);
-	
+
 	if (unicsv_datum_idx == DATUM_WGS84) {
 		lat = wpt->latitude;
 		lon = wpt->longitude;
@@ -1076,7 +1288,7 @@ unicsv_waypt_disp_cb(const waypoint *wpt)
 	}
 
 	gbfprintf(fout, "%d%s", unicsv_waypt_ct, unicsv_fieldsep);
-	
+
 	switch(unicsv_grid_idx) {
 
 	case grid_lat_lon_ddd:
@@ -1088,7 +1300,7 @@ unicsv_waypt_disp_cb(const waypoint *wpt)
 		cout = pretty_deg_format(lat, lon, 'm', unicsv_fieldsep, 0);
 		gbfputs(cout, fout);
 		break;
-		
+
 	case grid_lat_lon_dms: {
 		char *sep, *tmp;
 		cout = pretty_deg_format(lat, lon, 's', unicsv_fieldsep, 0);
@@ -1110,8 +1322,8 @@ unicsv_waypt_disp_cb(const waypoint *wpt)
 		if (! GPS_Math_WGS84_To_UKOSMap_M(wpt->latitude, wpt->longitude, &east, &north, map))
 			unicsv_fatal_outside(wpt);
 		gbfprintf(fout, "%s%s%5.0f%s%5.0f",
-			map, unicsv_fieldsep, 
-			east, unicsv_fieldsep, 
+			map, unicsv_fieldsep,
+			east, unicsv_fieldsep,
 			north);
 		break;
 	}
@@ -1124,26 +1336,27 @@ unicsv_waypt_disp_cb(const waypoint *wpt)
 			&east, &north, &zone, &zonec, unicsv_datum_idx))
 			unicsv_fatal_outside(wpt);
 		gbfprintf(fout, "%02d%s%c%s%.0f%s%.0f",
-			zone, unicsv_fieldsep, 
+			zone, unicsv_fieldsep,
 			zonec, unicsv_fieldsep,
-			east, unicsv_fieldsep, 
+			east, unicsv_fieldsep,
 			north);
 		break;
 	}
 	case grid_swiss: {
 		double north, east;
-		
-		if (! GPS_Math_WGS84_To_CH1903_NGEN(wpt->latitude, wpt->longitude, &east, &north))
+
+		if (! GPS_Math_WGS84_To_Swiss_EN(wpt->latitude, wpt->longitude, &east, &north))
 			unicsv_fatal_outside(wpt);
-		gbfprintf(fout, "%.f%s%.f%s",
-			east, unicsv_fieldsep, north, unicsv_fieldsep);
+		gbfprintf(fout, "%.f%s%.f",
+			east, unicsv_fieldsep, north);
+		break;
 
 	}
 	default:
-		gbfprintf(fout, "%.6f%s%.6f", lat, unicsv_fieldsep, lon);
+		gbfprintf(fout, "%.*f%s%.*f", llprec, lat, unicsv_fieldsep, llprec, lon);
 		break;
 	}
-	
+
 	if (cout) xfree(cout);
 
 	if FIELD_USED(fld_shortname) unicsv_print_str(shortname);
@@ -1241,7 +1454,7 @@ unicsv_waypt_disp_cb(const waypoint *wpt)
 			struct tm tm;
 			char buf[32];
 			time_t time = wpt->creation_time;
-			
+
 			if (opt_utc) {
 				time += atoi(opt_utc) * SECONDS_PER_HOUR;
 				tm = *gmtime(&time);
@@ -1260,7 +1473,7 @@ unicsv_waypt_disp_cb(const waypoint *wpt)
 			struct tm tm;
 			char buf[32], msec[12];
 			time_t time = wpt->creation_time;
-			
+
 			if (opt_utc) {
 				time += atoi(opt_utc) * SECONDS_PER_HOUR;
 				tm = *gmtime(&time);
@@ -1271,7 +1484,7 @@ unicsv_waypt_disp_cb(const waypoint *wpt)
 			if (wpt->microseconds > 0) {
 				int len = 6;
 				int ms = wpt->microseconds;
-				
+
 				while (len && (ms % 10 == 0)) {
 					ms /= 10;
 					len--;
@@ -1293,6 +1506,65 @@ unicsv_waypt_disp_cb(const waypoint *wpt)
 	if FIELD_USED(fld_garmin_state) unicsv_print_str(GMSD_GET(state, NULL));
 	if FIELD_USED(fld_garmin_country) unicsv_print_str(GMSD_GET(country, NULL));
 	if FIELD_USED(fld_garmin_phone_nr) unicsv_print_str(GMSD_GET(phone_nr, NULL));
+	if FIELD_USED(fld_garmin_phone_nr2) unicsv_print_str(GMSD_GET(phone_nr2, NULL));
+	if FIELD_USED(fld_garmin_fax_nr) unicsv_print_str(GMSD_GET(fax_nr, NULL));
+	if FIELD_USED(fld_garmin_email) unicsv_print_str(GMSD_GET(email, NULL));
+
+#ifdef UNICSV_GC_READY
+	if (waypt_empty_gc_data(wpt)) gc_data = NULL;
+	else gc_data = wpt->gc_data;
+
+	if FIELD_USED(fld_gc_id) {
+		gbfputs(unicsv_fieldsep, fout);
+		if (gc_data && gc_data->id) gbfprintf(fout, "%d", gc_data->id);
+	}
+	if FIELD_USED(fld_gc_type) {
+		if (gc_data) unicsv_print_str(gs_get_cachetype(gc_data->type));
+		else gbfputs(unicsv_fieldsep, fout);
+	}
+	if FIELD_USED(fld_gc_container) {
+		if (gc_data) unicsv_print_str(gs_get_container(gc_data->container));
+		else gbfputs(unicsv_fieldsep, fout);
+	}
+	if FIELD_USED(fld_gc_terr) {
+		gbfputs(unicsv_fieldsep, fout);
+		if (gc_data && gc_data->terr) gbfprintf(fout, "%.1f", (double)gc_data->terr / 10);
+	}
+	if FIELD_USED(fld_gc_diff) {
+		gbfputs(unicsv_fieldsep, fout);
+		if (gc_data && gc_data->diff) gbfprintf(fout, "%.1f", (double)gc_data->diff / 10);
+	}
+	if FIELD_USED(fld_gc_is_archived) {
+		if (gc_data && gc_data->is_archived) unicsv_print_str((gc_data->is_archived == status_true) ? "True" : "False");
+		else gbfputs(unicsv_fieldsep, fout);
+	}
+	if FIELD_USED(fld_gc_is_available) {
+		if (gc_data && gc_data->is_available) unicsv_print_str((gc_data->is_available == status_true) ? "True" : "False");
+		else gbfputs(unicsv_fieldsep, fout);
+	}
+	if FIELD_USED(fld_gc_exported) {
+		if (gc_data) unicsv_print_data_time(gc_data->exported);
+		else gbfputs(unicsv_fieldsep, fout);
+	}
+	if FIELD_USED(fld_gc_last_found) {
+		if (gc_data) unicsv_print_data_time(gc_data->last_found);
+		else gbfputs(unicsv_fieldsep, fout);
+	}
+	if FIELD_USED(fld_gc_placer) {
+		if (gc_data) unicsv_print_str(gc_data->placer);
+		else gbfputs(unicsv_fieldsep, fout);
+	}
+	if FIELD_USED(fld_gc_placer_id) {
+		gbfputs(unicsv_fieldsep, fout);
+		if (gc_data && gc_data->placer_id) gbfprintf(fout, "%d", gc_data->placer_id);
+	}
+	if FIELD_USED(fld_gc_hint) {
+		if (gc_data) unicsv_print_str(gc_data->hint);
+		else gbfputs(unicsv_fieldsep, fout);
+	}
+#endif
+	if (opt_format) unicsv_print_str(wpt->session->name);
+	if (opt_filename) unicsv_print_str(wpt->session->filename);
 
 	gbfputs(UNICSV_LINE_SEP, fout);
 }
@@ -1305,7 +1577,7 @@ unicsv_wr_init(const char *filename)
 {
 	fout = gbfopen(filename, "wb", MYNAME);
 
-	unicsv_outp_flags = 0;
+	memset(&unicsv_outp_flags, 0, sizeof(unicsv_outp_flags));
 	unicsv_grid_idx = grid_unknown;
 	unicsv_datum_idx = DATUM_WGS84;
 	unicsv_fieldsep = UNICSV_FIELD_SEP;
@@ -1313,7 +1585,7 @@ unicsv_wr_init(const char *filename)
 
 	if (opt_grid != NULL) {
 		int i;
-		
+
 		if (sscanf(opt_grid, "%d", &i)) {
 			unicsv_grid_idx = (grid_type) i;
 			if ((unicsv_grid_idx < GRID_INDEX_MIN) || (unicsv_grid_idx > GRID_INDEX_MAX))
@@ -1322,7 +1594,7 @@ unicsv_wr_init(const char *filename)
 		}
 		else unicsv_grid_idx = gt_lookup_grid_type(opt_grid, MYNAME);
 	}
-	
+
 	if (unicsv_grid_idx == grid_bng)
 		/* force datum to "Ord Srvy Grt Britn" / OSGB36 */
 		/* ! ignore parameter "Datum" ! */
@@ -1332,6 +1604,8 @@ unicsv_wr_init(const char *filename)
 		unicsv_datum_idx = DATUM_WGS84;	/* internal, becomes CH1903 */
 	else
 		unicsv_datum_idx = gt_lookup_datum_index(opt_datum, MYNAME);
+
+	llprec = atoi(opt_prec);
 }
 
 static void
@@ -1354,31 +1628,31 @@ unicsv_wr(void)
 		route_disp_all(NULL, NULL, unicsv_waypt_enum_cb);
 		break;
 	case posndata:
-		fatal(MYNAME ": Realtime positioning not supported.\n"); 
+		fatal(MYNAME ": Realtime positioning not supported.\n");
 	}
 
 	gbfprintf(fout, "No%s", unicsv_fieldsep);
 
 	switch(unicsv_grid_idx) {
-	case grid_bng: 
+	case grid_bng:
 /*		indexed parameters doesn't work under __win32__ (mingw)
 		gbfprintf(fout, "BNG-Zone%1$sBNG-East%1$sBNG-North", unicsv_fieldsep);
 */
 		gbfprintf(fout, "BNG-Zone%sBNG-East%sBNG-North",
 			unicsv_fieldsep, unicsv_fieldsep);
 		break;
-	case grid_utm: 
+	case grid_utm:
 /*		indexed parameters doesn't work under __win32__ (mingw)
 		gbfprintf(fout, "BNG-Zone%1$sBNG-East%1$sBNG-North", unicsv_fieldsep);
 */
 		gbfprintf(fout, "UTM-Zone%sUTM-Ch%sUTM-East%sUTM-North",
 			unicsv_fieldsep, unicsv_fieldsep, unicsv_fieldsep);
 		break;
-	case grid_swiss: 
+	case grid_swiss:
 		gbfprintf(fout, "Swiss-East%sSwiss-North",
 			unicsv_fieldsep);
 		break;
-	default: 
+	default:
 		gbfprintf(fout, "Latitude%sLongitude", unicsv_fieldsep);
 	}
 
@@ -1402,6 +1676,7 @@ unicsv_wr(void)
 	if FIELD_USED(fld_date) gbfprintf(fout, "%sDate", unicsv_fieldsep);
 	if FIELD_USED(fld_time) gbfprintf(fout, "%sTime", unicsv_fieldsep);
 	if FIELD_USED(fld_url) gbfprintf(fout, "%sURL", unicsv_fieldsep);
+
 	if FIELD_USED(fld_garmin_facility) gbfprintf(fout, "%sFacility", unicsv_fieldsep);
 	if FIELD_USED(fld_garmin_addr) gbfprintf(fout, "%sAddress", unicsv_fieldsep);
 	if FIELD_USED(fld_garmin_city) gbfprintf(fout, "%sCity", unicsv_fieldsep);
@@ -1409,7 +1684,27 @@ unicsv_wr(void)
 	if FIELD_USED(fld_garmin_state) gbfprintf(fout, "%sState", unicsv_fieldsep);
 	if FIELD_USED(fld_garmin_country) gbfprintf(fout, "%sCountry", unicsv_fieldsep);
 	if FIELD_USED(fld_garmin_phone_nr) gbfprintf(fout, "%sPhone", unicsv_fieldsep);
-	
+	if FIELD_USED(fld_garmin_phone_nr2) gbfprintf(fout, "%sPhone2", unicsv_fieldsep);
+	if FIELD_USED(fld_garmin_fax_nr) gbfprintf(fout, "%sFax", unicsv_fieldsep);
+	if FIELD_USED(fld_garmin_email) gbfprintf(fout, "%sEmail", unicsv_fieldsep);
+
+#ifdef UNICSV_GC_READY
+	if FIELD_USED(fld_gc_id) gbfprintf(fout, "%sGCID", unicsv_fieldsep);
+	if FIELD_USED(fld_gc_type) gbfprintf(fout, "%sType", unicsv_fieldsep);
+	if FIELD_USED(fld_gc_container) gbfprintf(fout, "%sContainer", unicsv_fieldsep);
+	if FIELD_USED(fld_gc_terr) gbfprintf(fout, "%sTerrain", unicsv_fieldsep);
+	if FIELD_USED(fld_gc_diff) gbfprintf(fout, "%sDifficulty", unicsv_fieldsep);
+	if FIELD_USED(fld_gc_is_archived) gbfprintf(fout, "%sArchived", unicsv_fieldsep);
+	if FIELD_USED(fld_gc_is_available) gbfprintf(fout, "%sAvailable", unicsv_fieldsep);
+	if FIELD_USED(fld_gc_exported) gbfprintf(fout, "%sExported", unicsv_fieldsep);
+	if FIELD_USED(fld_gc_last_found) gbfprintf(fout, "%sLast Found", unicsv_fieldsep);
+	if FIELD_USED(fld_gc_placer) gbfprintf(fout, "%sPlacer", unicsv_fieldsep);
+	if FIELD_USED(fld_gc_placer_id) gbfprintf(fout, "%sPlacer ID", unicsv_fieldsep);
+	if FIELD_USED(fld_gc_hint) gbfprintf(fout, "%sHint", unicsv_fieldsep);
+#endif
+	if (opt_format) gbfprintf(fout, "%sFormat", unicsv_fieldsep);
+	if (opt_filename) gbfprintf(fout, "%sFilename", unicsv_fieldsep);
+
 	gbfputs(UNICSV_LINE_SEP, fout);
 
 	switch(global_opts.objective) {
@@ -1434,7 +1729,7 @@ ff_vecs_t unicsv_vecs = {
 	FF_CAP_RW_ALL,
 	unicsv_rd_init,
 	unicsv_wr_init,
-	unicsv_rd_deinit, 
+	unicsv_rd_deinit,
 	unicsv_wr_deinit,
 	unicsv_rd,
 	unicsv_wr,
